from functools import cached_property
import requests
from typing import List, Tuple, Dict


def search_datasets_for_search_term(search_text: str) -> List[dict]:
    """
    Search the Renku Knowledge Graph for datasets that contain <search text>.
    All dataset fields are searched, including title, description, and keywords.

    Args:
        search_text: Text to search for in datasets.

    Returns: List of dictionaries describing each of the matching datasets found in the Renku Knowledge Graph.

    """
    r = requests.get('https://renkulab.io/knowledge-graph/datasets?query={}'.format(search_text))
    result_json = r.json()
    return result_json


def query_dataset_info(dataset_identifier: str) -> dict:
    """
    Query the Renku Knowledge Graph for all information about a dataset.
    Args:
        dataset_identifier: Renku Knowledge Graph dataset identifier

    Returns: Dictionary contianing the json describing the dataset.

    """
    r = requests.get('https://renkulab.io/knowledge-graph/datasets/{}'.format(dataset_identifier))
    result_json = r.json()
    return result_json


def query_project_info(project_path_id: str) -> dict:
    """
    Query for a project in the Renku Knowledge Graph

    Args:
        project_path_id: the path for a project

    Returns: dictionary describing the project.

    """
    r = requests.get('https://renkulab.io/knowledge-graph/projects/{}'.format(project_path_id))
    result_json = r.json()
    return result_json


def query_project_datasets(project_path_id: str) -> dict:
    """
    Query the Renku Knowledge Graph for all datasets related to a project.

    Args:
        project_path_id: the path for a project

    Returns: dictionary describing all of the datasets related to a project.

    """
    r = requests.get('https://renkulab.io/knowledge-graph/projects/{}/datasets'.format(project_path_id))
    result_json = r.json()
    return result_json


class RenkuDataSet:
    """Representation of a DataSet object in the Renku Knowledge Graph"""

    def __init__(self, data: dict):
        self.data = data
        self.validate(self.data)
        self.id = self.data['identifier']
        self.name = self.data['name']
        self.doi = self.data.get('sameAs', self.id)
        self.node_id = self.doi

    @classmethod
    def from_id(cls, identifier) -> 'RenkuDataSet':
        """Construct a RenkuDataSet object from a dataset identifier string"""
        if type(identifier) != str:
            raise TypeError(f'identifier must be of type str, received {type(identifier)}')

        data = query_dataset_info(identifier)

        if 'message' in data:
            if data['message'] == 'Resource not found':
                raise ValueError(f"Renku DataSet '{identifier}' not found")

        return cls(data)

    def __repr__(self):
        return f'RenkuDataSet(name: {self.name}, id: {self.id})'

    def to_dict(self):
        """Convert the RenkDataSet object to a dictionary for serialization."""
        return {'type': 'dataset', 'data': self.data}

    @classmethod
    def validate(cls, data: dict):
        """Make sure data is of type dict and contains the required keys."""
        if type(data) != dict:
            raise TypeError(f'data must be of type dict, received {type(data)}')

        required_fields = ['identifier', 'name', 'usedIn']
        for key in required_fields:
            if key not in data:
                raise ValueError(f'DataSet data missing key {key}')

    @cached_property
    def related_projects(self) -> List['RenkuProject']:
        """Query for all projects related to this dataset."""
        related_projects = []
        related_projects_list = self.data['usedIn']
        for p in related_projects_list:
            try:
                related_projects.append(RenkuProject.from_project_path_id(p['path']))
            except ValueError:
                print(f"Could not instantiate RenkuProject for id {p['path']}")
        return related_projects

    def to_cytoscape_node(self) -> dict:
        """Generate a data dictionary representing the dataset as a node in a cytoscape graph."""
        return {'data': {'id': self.doi, 'label': self.name}, 'classes': 'dataset'}

    def generate_related_network(self) -> Tuple[List[dict], Dict[str, 'RenkuProject']]:
        """Generate a data dictionary representing all of the projects related to this dataset in a cytoscape graph."""
        graph_elements = []
        node_metadata = {}
        for project in self.related_projects:
            graph_elements.append(project.to_cytoscape_node())
            graph_elements.append({'data': {'source': self.doi, 'target': project.id}})
            node_metadata[project.id] = project

        return graph_elements, node_metadata

    def generate_markdown_description(self):
        """Generate a description of this dataset formatted as markdown"""
        metadata_str = f"### DataSet '{self.name}'"
        metadata_str += f"\n\n**Identifier**: {self.id}"
        metadata_str += f"\n\n**DOI**: {self.doi}"
        metadata_dict = self.data
        if 'title' in metadata_dict:
            metadata_str += f"\n\n**Title:** {metadata_dict['title']}"
        if 'description' in metadata_dict:
            metadata_str += f"\n\n**Description:** {metadata_dict['description']}"
        if 'published' in metadata_dict:
            metadata_str += f"\n\n**Published by:** {metadata_dict['published']['creator'][0]['name']}"
        if 'created' in metadata_dict:
            metadata_str += f"\n\n**Created on:** {metadata_dict['created']}"

        metadata_str += f'\n\n**Used in {len(self.related_projects)} project(s)**'
        for project in self.related_projects:
            metadata_str += f'\n * {project.name}'

        return metadata_str


class RenkuProject:
    """Representation of a Project object in the Renku Knowledge Graph"""

    def __init__(self, data: dict):
        self.data = data
        self.validate(self.data)
        self.id = self.data['path']

        # I started with this as `name`, but there are many duplicates and it was confusing
        self.name = self.data['path']
        self.node_id = self.id

    @classmethod
    def from_project_path_id(cls, project_path_id) -> 'RenkuProject':
        """Construct a RenkuProject obkect from a project path."""
        if type(project_path_id) != str:
            raise TypeError(f'project_path_id must be of type str, received {type(project_path_id)}')

        data = query_project_info(project_path_id)

        if 'message' in data:
            if data['message'] == 'Resource not found':
                raise ValueError(f"Renku Project '{project_path_id}' not found")

        return cls(data)

    def __repr__(self):
        return f'RenkuProject(name: {self.name}, id: {self.id})'

    def to_dict(self):
        """Convert the RenkuProject object to a dictionary for serialization."""
        return {'type': 'project', 'data': self.data}

    @classmethod
    def validate(cls, data: dict):
        """Make sure data is of type dict and contains the required keys."""
        if type(data) != dict:
            raise TypeError(f'data must be of type dict, received {type(data)}')

        required_fields = ['path', 'name']
        for key in required_fields:
            if key not in data:
                raise ValueError(f'Project data missing key {key}')

    @cached_property
    def related_datasets(self) -> List[RenkuDataSet]:
        """Query for all datasets related to this project."""
        related_datasets = []
        related_datasets_list = query_project_datasets(self.id)
        for d in related_datasets_list:
            related_datasets.append(RenkuDataSet.from_id(d['identifier']))
        return related_datasets

    def to_cytoscape_node(self):
        """Generate a data dictionary representing the project as a node in a cyctoscape graph."""
        return {'data': {'id': self.id, 'label': self.name}, 'classes': 'project'}

    def generate_related_network(self) -> Tuple[List[dict], Dict[str, RenkuDataSet]]:
        """Generate a data dictionary representing the project as a node in a cyctoscape graph."""
        graph_elements = []
        node_metadata = {}
        for dataset in self.related_datasets:
            graph_elements.append(dataset.to_cytoscape_node())
            graph_elements.append({'data': {'source': self.id, 'target': dataset.doi}})
            node_metadata[dataset.id] = dataset

        return graph_elements, node_metadata

    def generate_markdown_description(self):
        """Generate a description of this dataset formatted as markdown"""
        metadata_str = f"### Project '{self.name}'"
        metadata_str += f"\n\n**Identifier**: {self.id}"
        metadata_dict = self.data
        if 'description' in metadata_dict:
            metadata_str += f"\n\n**Description:** {metadata_dict['description']}"
        if 'published' in metadata_dict:
            metadata_str += f"\n\n**Published by:** {metadata_dict['published']['creator'][0]['name']}"
        if 'created' in metadata_dict:
            metadata_str += f"\n\n**Created on:** {metadata_dict['created']}"
        if 'urls' in metadata_dict:
            if 'web' in metadata_dict['urls']:
                metadata_str += f"\n\n**Project Page:** {metadata_dict['urls']['web']}"

        metadata_str += f'\n\n**Uses {len(self.related_datasets)} datasets**'
        for ds in self.related_datasets:
            metadata_str += f'\n * {ds.name}'

        return metadata_str
